""" Twisted Asyncronous Database Api

TODO:

  * remaining blocking calls (those that use 'maybeDeferred')
        should be truly deferred and queued inplace within the
        highest threadsafty context that dbapi allows.

  * connection pooling

"""

from twisted.python import reflect,failure
from twisted.internet import defer,threads,reactor
import sys,Queue

def maybeDeferred(func):
    def wrapper(*args,**kwargs):
        return defer.maybeDeferred(func,*args,**kwargs)
    return wrapper

def deferToThread(func):
    def wrapper(*args,**kwargs):
        return threads.deferToThread(func(*args,**kwargs))
    return wrapper

def deferToThreadQueue(func,threadQueue):
    def wrapper(*args,**kwargs):
        d = defer.Deferred()
        threadQueue.put((d,func,args,kwargs))
        return d
    return wrapper

class ThreadedQueue(object):
    def __init__(self):
        self.queue = Queue.Queue()
        self.alive = True
        self.busy = True
    def __del__(self):
        def stillBusy(selfToo):
            pass
        if self.busy:
            reactor.callLater(0,stillBusy,self)
    def __die__(self):
        self.alive = False
    def die(self):
        d = defer.Deferred()
        self.put((d,self.__die__,[],{}))
        return defer.DeferredList([d,self.threadRun])
    def put(self,*args,**kwargs):
        self.queue.put(*args,**kwargs)
    def run(self):
        def runComplete(selfToo,x):
            self.busy = False
        def threadLoop():
            self.busy = True
            while self.alive:
                (deferred,func,args,kwargs) = self.queue.get()
                try:
                    result = func(*args,**kwargs)
                except:
                    exc_info=sys.exc_info()
                    deferred.errback(failure.Failure(*exc_info))
                else:
                    deferred.callback(result)
        self.threadRun = threads.deferToThread(threadLoop)
        return self.threadRun.addCallback(lambda _: runComplete(self,_))

class Cursor(object):
    def __init__(self,connection,cursor,threadedQueue=None):
        self.__tada_connection__ = connection
        self.__tada_cursor__ = cursor
        self.__tada_threadQueue__ = threadedQueue
    def __del__(self):
        def waitforqueue(selfToo):
            pass
        if self.__tada_threadQueue__ is not None and self.__tada_threadQueue__ != self.__tada_connection__.__tada_threadQueue__ and self.__tada_threadQueue__.alive:
            self.__tada_threadQueue__.die().addCallback(lambda _: waitforqueue(self))
    def __getattr__(self,name):
        attr = getattr(self.__dict__["__tada_cursor__"],name)
        if callable(attr):
            if self.__tada_connection__.__tada_module__.threadsafety < 3:
                if self.__tada_threadQueue__ is not None:
                    return deferToThreadQueue(attr,self.__tada_threadQueue__)
                else:
                    return maybeDeferred(attr)
            else:
                return deferToThread(attr)
        return attr
    def __setattr__(self,name,value):
        if name.startswith("__tada"):
            self.__dict__[name]=value
        else:
            self.__dict__["__tada_cursor__"].name = value
    def __repr__(self):    
        return "<tada.Cursor instance of %s>"%self.__tada_cursor__

class Connection(object):
    def __init__(self,module,connection,threadQueue=None):
        self.__tada_module__ = module
        self.__tada_connection__ = connection
        self.__tada_threadQueue__ = threadQueue
        self.__tada_threadRun__ = None
    def __tada_cursor__(self):
        if self.__tada_module__.threadsafety < 2:
            if self.__tada_threadQueue__ is not None:
                d=deferToThreadQueue(self.__tada_connection__.cursor,self.__tada_threadQueue__)()
                d.addCallback(lambda cursor: Cursor(self,cursor,self.__tada_threadQueue__))
                return d
            else:
                return defer.maybeDeferred(lambda: Cursor(self,self.__tada_connection__.cursor(),self.__tada_threadQueue__))
        else:
            threadedQueue = ThreadedQueue()
            self.__tada_threadRun__ = threadedQueue.run()
            self.__tada_threadRun__.addCallback(lambda x: x)
            d = defer.Deferred()
            threadedQueue.put((d,self.__tada_connection__.cursor,[],{}))
            d.addCallbacks(lambda cursor: Cursor(self,cursor,threadedQueue),lambda failure:threadedQueue.die())
            return d
    def __del__(self):
        def waitforqueue(selfToo): pass
        if self.__tada_threadQueue__ is not None and self.__tada_threadQueue__.alive:
            self.commit()
            self.__tada_threadQueue__.die().addCallback(lambda _:waitforqueue(self))
    def __getattr__(self,name):
        if name == "cursor": return self.__tada_cursor__
        attr = getattr(self.__dict__["__tada_connection__"],name)
        if callable(attr):
            if self.__tada_threadQueue__ is not None:
                return deferToThreadQueue(attr,self.__tada_threadQueue__)
            else:
                return maybeDeferred(attr)
        return attr
    def __setattr__(self,name,value):
        if name.startswith("__tada"):
            self.__dict__[name]=value
        else:
            self.__dict__["__tada_connection__"].name = value
    def __repr__(self):    
        return "<tada.Connection instance of %s>"%self.__tada_connection__

class Tada(object):
    def __init__(self,dbapi):
        self.__tada_module__ = reflect.namedModule(dbapi)
        assert self.__tada_module__.apilevel=='2.0',"Incompatible apilevel.  Expected 2.0, got %s"%self.__tada_module__.apilevel
    def __tada_connect__(self,*args,**kwargs):
        def connect():
            connection = self.__tada_module__.connect(*args,**kwargs)
            return Connection(self,connection)
        if self.threadsafety < 1:
            return maybeDeferred(connect)()
        else:
            threadQueue = ThreadedQueue()
            threadQueue.run()
            d = defer.Deferred()
            threadQueue.put((d,self.__tada_module__.connect,args,kwargs))
            def badConnect(failure):
                return threadQueue.die().addBoth(lambda x: failure)
            d.addCallbacks(lambda connection: Connection(self,connection,threadQueue),badConnect)
            return d
    def __getattr__(self,name):
        if name == "connect": return self.__tada_connect__
        attr = getattr(self.__dict__["__tada_module__"],name)
        if callable(attr): return maybeDeferred(attr)
        return attr
    def __setattr__(self,name,value):
        if name.startswith("__tada"):
            self.__dict__[name]=value
        else:
            self.__dict__["__tada_module__"].name = value
    def __repr__(self):    
        return "<tada.Tada instance of %s>"%self.__tada_module__

class Tadah(Tada):
    def __init__(self,dbapi,*connargs,**connkwargs):
        self.__tada_connargs__ = connargs
        self.__tada_connkwargs__ = connkwargs
        Tada.__init__(self,dbapi)
    def connect(self):
        return self.__tada_connect__(*self.__tada_connargs__,**self.__tada_connkwargs__)
    def cursor(self):
        def gotConnection(connection):
            return connection.cursor()
        return self.connect().addCallback(gotConnection)
    def execute(self,query,*args):
        def gotCursor(cursor):
            def gotResults(results):
                return cursor
            return cursor.execute(query,*args).addCallback(gotResults)
        return self.cursor().addCallback(gotCursor)
    def __repr__(self):
        return "<tada.Tadah instance of %s>"%self.__tada_module__

