import types
from copy import copy
from twisted.internet import defer
from twisted.python import log as twistedlog

from util import genAside

def log(message):
    twistedlog.debug("[tada.tao] "+message)

class PrimaryKeyMissing(Exception):
    pass

class SQLString(str):
    def __or__(self,other):
        return SQLString('('+str(self)+' OR '+str(other)+')')
    def __and__(self,other):
        return SQLString('('+str(self)+' AND '+str(other)+')')
    def __invert__(self):
        return SQLString('(NOT '+str(self)+')')

def fmt(func):
    def wrapper(self,value):
        if isinstance(value,SQLString):
            return value
        return func(self,value)
    return wrapper

class Col(object):
    __name__  = None
    default = None
    value = None
    length = None
    null = False
    table = None
    pkey = False
    create = True

    def __init__(self,name,**kwargs):
        self.name = name
        for key,value in kwargs.items():
            if hasattr(self,key): setattr(self,key,value)

    def _get_name(self):
        return self.__name__
    def _set_name(self,value):
        assert self.__name__ is None, "Cannot change column name. Requested from %s to %s."%(self.__name__,value)
        self.__name__ = value
    name = property(_get_name,_set_name)

    def _get_type(self):
        return types.NoneType
    type = property(_get_type)

    def _get_dbtype(self):
        return 'None'
    dbtype = property(_get_dbtype)

    def __str__(self):
        name = self.name

        # TODO dbapi checks here:

        return SQLString(name)

    def _get_fullname(self):
        if self.table is not None: return SQLString(str(self.table)+'.'+str(self))
        return str(self)
    fullname=property(_get_fullname)

    def _get_fullpyname(self):
        if self.table is not None: return SQLString(str(self.table)+'.'+str(self.pyname))
        return str(self)
    fullpyname=property(_get_fullpyname)

    def __repr__(self):
        result = self.dbtype
        if not self.null: result += ' NOT NULL'
        if self.default is not None: result += " DEFAULT %s"%self.format(self.default)
        return result

    def wrap(self,other):
        if type(other)==str or type(other)==unicode:
            return "'"+other+"'"
        return other

    @fmt
    def format(self,value):
        return str(value)

    def __lt__(self,other):
        if isinstance(other,Col): other = other.fullname
        return SQLString('( '+ self.fullname + ' < ' + str(self.wrap(other)) + ' )')
    def __le__(self,other):
        if isinstance(other,Col): other = other.fullname
        return SQLString('( '+ self.fullname + ' <= ' + str(self.wrap(other)) + ' )')
    def __eq__(self,other):
        if isinstance(other,Col): other = other.fullname
        if other is None:
            inner = ' IS NULL'
        else:
            inner = ' = ' + str(self.wrap(other))
        return SQLString('( '+ self.fullname + inner + ' )')
    def __ne__(self,other):
        if isinstance(other,Col): other = other.fullname
        return SQLString('( NOT ' + self.fullname + ' = ' + str(self.wrap(other)) + ' )')
    def __gt__(self,other):
        if isinstance(other,Col): other = other.fullname
        return SQLString('( '+ self.fullname + ' > ' + str(self.wrap(other)) + ' )')
    def __ge__(self,other):
        if isinstance(other,Col): other = other.fullname
        return SQLString('( '+ self.fullname + ' >= ' + str(self.wrap(other)) + ' )')
    def __xor__(self,other):
        if isinstance(other,Col): other = other.fullname
        return SQLString('('+ self.fullname +' ILIKE '+str(self.wrap(other))+')')
    def __or__(self,other):
        if isinstance(other,Col): other = other.fullname
        return SQLString('('+ self.fullname +' LIKE '+str(self.wrap(other))+')')

    def __pos__(self):
        return SQLString(self.fullname+' ASC')
    def __neg__(self):
        return SQLString(self.fullname+' DESC')

class StringCol(Col):
    char = False
    
    def _get_type(self):
        return types.StringType
    type = property(_get_type)

    def _get_dbtype(self):
        if self.length is not None:
            if self.char:
                return 'char(%d)'%self.length
            else:
                return 'varchar(%d)'%self.length
        return 'text'
    dbtype = property(_get_dbtype)
    
    @fmt
    def format(self,value):
        # TODO this should check dbapi-type and deal with it accordingly
        # if self.table.__tadah__.__tada_module__.__name__ == ""
        return repr(str(value))

class IntCol(Col):
    def _get_type(self):
        return types.IntType
    type = property(_get_type)

    def _get_dbtype(self):
        if self.length is not None: return 'int(%d)'%self.length
        return 'int'
    dbtype = property(_get_dbtype)

class TimestampCol(StringCol):
    def _get_type(self):
        return types.IntType
    type = property(_get_type)

    def _get_dbtype(self):
        return 'TIMESTAMP'
    dbtype = property(_get_dbtype)

class DateCol(TimestampCol):
    def _get_dbtype(self):
        return 'DATE'
    dbtype = property(_get_dbtype)

class Row(object):
    def __all__(self):
        all=[]
        for attrName in filter(lambda x: not x.startswith('_'),dir(self)):
            attr = getattr(self,attrName)
            if not callable(attr): all.append(attr)
        return all

class ResultRow(Row):
    pass

class TableRow(Row):
    def __init__(self):
        Row.__init__(self)
        self.clean()
    def clean(self):
        self.__pristine__ = [copy(getattr(self,x)) for x in filter(lambda x: not callable(getattr(self,x)),filter(lambda x: not x.startswith('_'),dir(self)))]
    def dirty(self):
        columns=[]
        for pristine,workingcopy in zip(self.__pristine__,map(lambda x: getattr(self,x),filter(lambda x: not callable(getattr(self,x)),filter(lambda x: not x.startswith('_'),dir(self))))):
            if pristine.value != workingcopy.value: columns.append(workingcopy)
        return columns
    def revert(self):
        for pristine,workingcopy in zip(self.__pristine__,filter(lambda x: not callable(getattr(self,x)),filter(lambda x: not x.startswith('_'),dir(self)))):
            workingcopy.value=pristine.value
    def update(self):
        dirtyList = self.dirty()
        if len(dirtyList) == 0: return defer.maybeDeferred(lambda :True)
        pkeys=[]
        for column in map(lambda x: getattr(self,x),filter(lambda x: not callable(getattr(self,x)),filter(lambda x: not x.startswith('_'),dir(self)))):
            if column.pkey:
                pkeys.append(column)
        if len(pkeys)==0:
            raise PrimaryKeyMissing, "Table `%s` has no allocated primary key."%str(self.__table__)
        def updated(cursor):
            connection = cursor.__tada_connection__
            del cursor
            self.clean()
            return connection.commit().addCallback(lambda _: True)
        sql = "UPDATE %s SET %s WHERE %s"%(str(self.__table__),",".join(map(lambda x: "%s=%s"%x,zip(map(lambda x: x.name,dirtyList),map(lambda x: x.format(x.value),dirtyList))))," AND ".join(map(lambda x: "%s=%s"%x,zip(map(lambda x: x.name,pkeys),map(lambda x: x.format(x.value),pkeys)))))
        log(sql)
        return self.__table__.__tadah__.execute(sql).addCallback(updated)
    def delete(self):
        def deleted(cursor):
            connection = cursor.__tada_connection__
            del cursor
            return connection.commit().addCallback(lambda _: True)
        pkeys=[]
        for column in map(lambda x: getattr(self,x),filter(lambda x: not callable(getattr(self,x)),filter(lambda x: not x.startswith('_'),dir(self)))):
            if column.pkey:
                pkeys.append(column)
        if len(pkeys)==0:
            raise PrimaryKeyMissing, "Table `%s` has no allocated primary key."%str(self.__table__)
        sql = "DELETE FROM %s WHERE %s"%(str(self.__table__)," AND ".join(map(lambda x: "%s=%s"%x,zip(map(lambda x: x.name,pkeys),map(lambda x: x.value,pkeys)))))
        log(sql)
        return self.__table__.__tadah__.execute(sql).addCallback(deleted)
    def __del__(self):
        self.update().addCallback(lambda _: True)

class Table(Row):
    __tadah__ = None
    __table__ = None

    def __init__(self,tadah=None,**kwargs):
        if tadah is not None: self.__tadah__ = tadah
        for key,value in kwargs.items():
            if hasattr(self,key): setattr(self,key,value)
        for attrName in filter(lambda x: not x.startswith('_'),dir(self)):
            attr = getattr(self,attrName)
            if not callable(attr):
                attr.table = self
                attr.pyname = attrName

    def __str__(self):
        return self.__table__

    def __repr__(self):
        result = 'CREATE TABLE '+self.__table__+' (\n'
        for attrName in filter(lambda x: not x.startswith('_'),dir(self)):
            attr = getattr(self,attrName)
            if not callable(attr) and attr.create: result += str(attr) + ' ' + repr(attr) + ',\n'
        if result[-2:]==',\n': result=result[:-2]
        result += ');'
        return result;

    def insert(self,**kwargs):
        def result(cursor):
            connection = cursor.__tada_connection__
            try:
                lastrowid = cursor.lastrowid
            except AttributeError:
                lastrowid = None
            del cursor
            return connection.commit().addCallback(lambda _: lastrowid)
        columns={}
        for columnname,value in kwargs.items():
            columns[columnname]=getattr(self,columnname).format(value)
        sql =  "INSERT INTO %s (%s) VALUES (%s)"%(self.__table__,','.join(map(lambda x:getattr(self,x).name,columns.keys())),','.join(map(lambda x:str(x),columns.values())))   
        log(sql)
        return self.__tadah__.execute(sql).addCallback(result)

    def select(self,*args,**kwargs):
        where = SQLString('')
        columns = []
        limit = None
        order = []
        tables = [self]
        for arg in args:
            if isinstance(arg,Col):
                columns.append(arg)
                tables.append(arg.table)
            if type(arg) == SQLString:
                if len(where) == 0:
                    where = arg
        tabledict=dict([(x.__table__,x) for x in tables])
        tables=tabledict.values()
        if len(tables)==1: columns+=filter(lambda x: ((not callable(x)) and (x.pkey)),map(lambda x: getattr(self,x),filter(lambda x: not x.startswith('_'),dir(self))))
        columns=list(set(columns)) # XXX is this the best-way(tm) ?
        for kw,value in kwargs.items():
            if kw=='limit': limit = value
            if kw=='order': order = value
            pass
        if len(columns) == 0: columns=self.__all__()
        select = "SELECT %s FROM %s"%(','.join(map(lambda x: x.fullname,columns)),",".join(map(lambda x: x.__table__,tables)))
        if len(where): select+=" WHERE %s"%where
        if len(order)>0: select+=' ORDER BY %s'%','.join(map(lambda x: str(x),order))
        if limit is not None: select+=' LIMIT %d'%limit
        def result(cursor):
            self.cursor = cursor
            def gotRows(rows):
                del self.cursor
                allRows = []
                for enum,row in enumerate(rows):
                    if len(tables)==1:
                        attributes = dict([(col.pyname,copy(col)) for col in columns])
                        attributes['__table__']=self
                        allRows.append(type('TableRow',(TableRow,),attributes)())
                        for attr,value in zip(map(lambda x:x.pyname,columns),row):
                            getattr(allRows[-1],attr).value = value
                        allRows[-1].clean()    
                    else:
                        attributes = dict([(col.fullpyname.replace('.','_'),copy(col)) for col in columns])
                        attributes['__table__']=self
                        allRows.append(type('ResultRow',(ResultRow,),attributes)())
                        for attr,value in zip(map(lambda x:x.fullpyname.replace('.','_'),columns),row):
                            getattr(allRows[-1],attr).value = value
                return allRows
            return cursor.fetchall().addCallback(gotRows)
        log(select)
        return self.__tadah__.execute(select).addCallback(result)

    @genAside
    def create(gen,self):
        sql=repr(self)
        log(sql)
        yield self.__tadah__.execute(sql)
        cursor = gen.result
        connection = cursor.__tada_connection__
        del cursor
        yield connection.commit(); gen.result
        del connection
        gen._return(True)

    def drop(self):
        sql="DROP TABLE "+str(self)
        log(sql)
        return self.__tadah__.execute(sql)

